KNOWN_LICENSES = [
    "Aladdin Free Public License (AFPL)", "CC0 1.0 Universal (CC0 1.0) Public Domain Dedication",
    "DFSG approved", "Eiffel Forum License (EFL)", "Free For Educational Use", "Free For Home Use",
    "Free for non-commercial use", "Freely Distributable", "Free To Use But Restricted", "Freeware",
    "Netscape Public License (NPL)", "Nokia Open Source License (NOKOS)", "Academic Free License (AFL)",
    "Apache Software License", "Apple Public Source License", "Artistic License",
    "Attribution Assurance License", "BSD License", "CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
    "Common Public License", "Eiffel Forum License", "European Union Public Licence 1.0 (EUPL 1.0)",
    "European Union Public Licence 1.1 (EUPL 1.1)", "GNU Affero General Public License v3",
    "GNU Affero General Public License v3 or later (AGPLv3+)", "GNU Free Documentation License (FDL)",
    "GNU General Public License (GPL)", "GNU General Public License v2 (GPLv2)",
    "GNU General Public License v2 or later (GPLv2+)", "GNU General Public License v3 (GPLv3)",
    "GNU General Public License v3 or later (GPLv3+)", "GNU Lesser General Public License v2 (LGPLv2)",
    "GNU Lesser General Public License v2 or later (LGPLv2+)", "GNU Lesser General Public License v3 (LGPLv3)",
    "GNU Lesser General Public License v3 or later (LGPLv3+)", "GNU Library or Lesser General Public License (LGPL)",
    "IBM Public License", "Intel Open Source License", "ISC License (ISCL)", "Jabber Open Source License",
    "MIT License", "MITRE Collaborative Virtual Workspace License (CVW)", "Motosoto License",
    "Mozilla Public License 1.0 (MPL)", "Mozilla Public License 1.1 (MPL 1.1)", "Mozilla Public License 2.0 (MPL 2.0)",
    "Nethack General Public License", "Nokia Open Source License", "Open Group Test Suite License",
    "Python License (CNRI Python License)", "Python Software Foundation License", "Qt Public License (QPL)",
    "Ricoh Source Code Public License", "Sleepycat License", "Sun Industry Standards Source License (SISSL)",
    "Sun Public License", "University of Illinois/NCSA Open Source License", "Vovida Software License 1.0",
    "W3C License", "X.Net License", "zlib/libpng License", "Zope Public License", "Other/Proprietary License",
    "Public Domain", "Repoze Public License", ]
